
#include <string>
#include <iostream>
#include <fstream>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/split.hpp>

#include "MiniGameState.H"
#include "MiniTacticalGameState.H"
#include "Player.H"
#include "RLComp08Bot1.H"
#include "RLComp08Bot2.H"
#include "RLComp08Bot3.H"
#include "SDL_GUI.H"
#include "SDL_init.H"
#include "Simulation.H"

using namespace std; 

static bool debug = false;


Simulation::Simulation()
{
  trace = NULL; 
  parmsfile = NULL; 
  enable_gui = false; 
  gui_delay = 25; 
  winner = -1; 
  timestep = 0; 
  max_timesteps = -1; 
  
  player0Ptr = NULL;
  player1Ptr = NULL; 
  parmsPtr = NULL; 
  statePtr = NULL; 
  
  // FIXME: Assuming 2 players
  statsPtr = new SimStats; 
  statsPtr->set_numplayers(2); 
  
  players_fullview = false;
  initial_wait = false;   
}

Simulation::~Simulation()
{
  DELETE(trace);
  DELETE(player0Ptr);
  DELETE(player1Ptr);
  DELETE(parmsPtr);
  DELETE(statePtr);
  DELETE(statsPtr);
}


void Simulation::open_trace(char * filename)
{
  cout << "Reading in trace file ... " << endl;
  
  ifstream inputStream;
  inputStream.open(filename);
  if( !inputStream ) {
    cerr << "Error opening trace file" << endl;
    exit(-1);
  }          
  
  string str; 
  getline(inputStream, str); 
  boost::replace_all(str, "$", "=");
  
  trace = new vector<string>;
  
  boost::split(*trace, str, boost::is_any_of("@"));

  long seed = to_long((*trace)[0]);
  srand(seed); 
}

void Simulation::create_state()
{
  statePtr = new MiniGameState; 
}

void Simulation::setup_players()
{  
  player0Ptr = new RLComp08Bot2(0);
  //player0Ptr = new TestPlayer(0);
  //player0Ptr = new TestPlayer(0);
  player0Ptr->set_parms(parmsPtr);
  //player0Ptr->set_fullview(players_fullview); 
  player0Ptr->set_fullview(true); 
  
  player1Ptr = new RLComp08Bot3(1);
  //player1Ptr = new TestPlayer(1);
  player1Ptr->set_parms(parmsPtr);
  //player1Ptr->set_fullview(players_fullview); 
  player1Ptr->set_fullview(true); 
  
  statsPtr->set_numplayers(2); 
}

void Simulation::init_parms()
{
  parmsPtr = new MiniGameParameters;
  if (parmsfile != NULL)
    load_parms(parmsfile, parmsPtr);
  
  if (trace != NULL)
  {
    // trace->[1] contains the parms
    parmsPtr->deserialize((*trace)[1]);
  }
  
}

void Simulation::init_state()
{
  if (trace != NULL)
  {
    statePtr->init(*parmsPtr, true);   // delay object setup
    
    // trace->[2] contains the initial state
    statePtr->from_string((*trace)[2]); 
  }  
  else
  {
    statePtr->init(*parmsPtr, false);
  }    
}

void Simulation::pre_simloop()
{
}

void Simulation::post_simloop()
{
}

void Simulation::pre_simiter()
{
}

void Simulation::post_simiter()
{
}

void Simulation::play() 
{
  cout << "Setting up .. " << endl;
  
  create_state();
  init_parms(); 
  setup_players(); 
  init_state();  

  unsigned int trace_index = 3; 

  cout << "Initializing graphics .. " << endl; 

  SDL_GUI<MiniGameState> gui;
  
  if (enable_gui) {

    SDL_init::video_init();
    
    std::map<std::string, SDL_GUI<MiniGameState>::Marker> markers;
    markers["worker"] = SDL_GUI<MiniGameState>::MARKER_H; // mark workers
  
    gui.init(1024, 768, *statePtr, markers);
    gui.display();
  }
    
  // compute initial views
  cout << "Computing initial views ... " << endl; 
  
  boost::array<std::string, MiniGameState::PLAYER_NUM> views;

  FORS (i, MiniGameState::PLAYER_NUM) {
    ostringstream os;
    statePtr->encode_view(i, os);
    views[i] = os.str();
  }

  // send them ...

  // run game
  
  cout << "Running the game ... " << endl; 

  winner = -1; 

  boost::array<std::string, MiniGameState::PLAYER_NUM> acts;
  timestep = 0; 
  bool out_of_time = false;
  
  pre_simloop();
    
  do {
        
    if (max_timesteps > 0 && timestep >= max_timesteps) {
      out_of_time = true; 
      break;
    }
    
    // needed for xvidcap when taking screenshots of initial state
    if (initial_wait && timestep == 1) sleep(7); 
    
    pre_simiter();        
    
    if (debug)
    {
      cout << endl << "## Time step " << timestep << endl;
      boost::array<std::string, MiniGameState::PLAYER_NUM> views;
    }
  
    statePtr->simulation_step(acts, views);

    if (enable_gui)
    {
      gui.event();
      gui.display();
    }

    // send views + receive actions ...

    if (trace == NULL)
    {
      if (players_fullview)
      {
        player0Ptr->set_state(statePtr);         
        player1Ptr->set_state(statePtr); 
      }
      else 
      {
        if (player0Ptr->get_fullview())
          player0Ptr->set_state(statePtr);         
        
        if (player1Ptr->get_fullview())
          player1Ptr->set_state(statePtr); 
      }
              
      acts[0] = player0Ptr->receive_actions(views[0]);
      acts[1] = player1Ptr->receive_actions(views[1]);        
    }
    else 
    {
      if (trace_index >= trace->size())
        break; 
      
      acts[0] = (*trace)[trace_index++];
      acts[1] = (*trace)[trace_index++];
    }
      
    post_simiter();    

    if (enable_gui)
      gui.delay(gui_delay);

    winner = statePtr->check_win(); 
    
    timestep++;    
    
  } while (!gui.quit && winner < 0);
  
  gui.quit = true; 

  cout << "Timesteps = " << timestep;
  cout << (out_of_time ? " (max reached). " : ". ");
  
  if (winner >= 0) 
    cout << " Winner is player " << winner << endl;
  else
    cout << " No winner" << endl;
  
  post_simloop();   
  
  collect_endgame_stats();  
}

void Simulation::collect_endgame_stats()
{
  // Add 1 to the number of simulations
  statsPtr->inc_num();
  
  if (winner >= 0 && winner < statsPtr->get_numplayers())
    statsPtr->inc_win(winner);
  
  FORALL(statePtr->all_objs, iter)
  {
    GameObj<MiniGameState> * objPtr = *iter; 
    
    // when owner > num_players, it is not owned by either (eg. mp)
    if (!objPtr->dead() && objPtr->owner < statsPtr->get_numplayers())
    {
      statsPtr->add_remaining_units(objPtr->owner, 1);
      statsPtr->add_remaining_ttlhp(objPtr->owner, objPtr->hp);
    }
  }
}

